from functools import partial
from operator import itemgetter
from typing import Callable, TypeVar

from .Either import Either, success, fail
from .Void import Void, VoidType

U = TypeVar('U')
V = TypeVar('V')

type Position = int
type ErrorType = str

type ParseResultSucceeded[U] = tuple[Position, U]
type ParseResultFailed = tuple[Position, ErrorType]
type ParseResult[U] = Either[ParseResultSucceeded[U], ParseResultFailed]

type ParserSucceeded[U] = Callable[[str, Position], ParseResultSucceeded[U]]
type ParserFailed = Callable[[str, Position], ParseResultFailed]
type Parser[U] = Callable[[str, Position], ParseResult[U]]

def satisfy(predicate: Callable[[str], bool], desc: str) -> Parser[str]:
    def f(s: str, idx: Position) -> ParseResult[str]:
        if idx == len(s):
            return fail((idx, desc))
        c = s[idx]
        if not predicate(c):
            return fail((idx, desc))
        idx += 1
        return success((idx, c))
    return f

def many(parser: Parser[str]) -> Parser[list[str]]:
    def f(s: str, idx: Position) -> ParseResult[list[str]]:
        out = []
        while True:
            res = parser(s, idx)
            if not res.is_success():
                return success((idx, out))
            idx, token = res.get()
            out.append(token)
    return f

def many1(parser: Parser[str]) -> Parser[list[str]]:
    def f(s: str, idx: Position) -> ParseResult[list[str]]:
        out = []

        res = parser(s, idx)
        if not res.is_success():
            return res
        idx, token = res.get()
        out.append(token)

        res = many(parser)(s, idx)
        if res.is_success():
            idx, token = res.get()
            out.extend(token)

        return success((idx, out))
    return f

def literal(s: str, *, desc: str | None = None) -> Parser[str]:
    if desc is None:
        desc = f'string "{s}"'
    len_literal = len(s)
    def f(s_: str, idx: Position) -> ParseResult[str]:
        if idx > len(s_) - len_literal or s_[idx:idx+len_literal] != s:
            return fail((idx, desc))
        idx += len_literal
        return success((idx, s))
    return f

def parse_eof(s: str, idx: Position) -> ParseResult[VoidType]:
    if idx != len(s):
        return fail((idx, 'eof'))
    return success((idx, Void))

def with_default(parser: Parser[U], *, default: U) -> ParserSucceeded[U]:
    def f(s: str, idx: Position) -> ParseResult[U]:
        res = parser(s, idx)
        if res.is_success():
            return res
        return success((idx, default))
    return f

def with_description(parser: Parser[U], desc: str) -> Parser[U]:
    def f(s: str, idx: Position) -> ParseResult[U]:
        res = parser(s, idx)
        if res.is_success():
            return res
        idx, _ = res.get()
        return fail((idx, desc))
    return f

def pmap(func: Callable[[U], V], parser: Parser[U]) -> Parser[V]:
    def f(s: str, idx: Position) -> ParseResult[V]:
        res = parser(s, idx)
        if not res.is_success():
            return res
        idx, token = res.get()
        return success((idx, func(token)))
    return f

def pchain(*parsers: Parser[U]) -> Parser[list[U]]:
    def f(s: str, idx: Position) -> ParseResult[list[U]]:
        out = []

        for parser in parsers:
            res = parser(s, idx)
            if not res.is_success():
                return res
            idx, val = res.get()
            if val is not Void:
                out.append(val)

        return success((idx, out))
    return f

join: Callable[[str], str] = lambda xs: ''.join(xs)
pjoin = partial(pmap, join)

def _join_descriptions(xs: list[str]) -> str:
    if len(xs) == 0:
        return ''
    if len(xs) == 1:
        return xs[0]
    *init, tail = xs
    return ', '.join(init) + ' or ' + tail

def anyof(*parsers: Parser[U]) -> Parser[U]:
    def f(s: str, idx: Position) -> ParseResult[U]:
        descriptions = []

        for parser in parsers:
            res = parser(s, idx)
            if res.is_success():
                return res
            _, desc = res.get()
            descriptions.append(desc)

        return fail((idx, _join_descriptions(descriptions)))
    return f

pselect = lambda idx, parser: pmap(itemgetter(idx), parser)

sepby1: Callable[[Parser[U], Parser[V]], Parser[U]] = lambda parser, separator: pmap(
    lambda xs: [xs[0], *xs[1]],
    pchain(
        parser,
        many(pselect(1, pchain(separator, parser))),
    ),
)

const: Callable[[U], Callable[[V], U]] = lambda x: lambda _: x
pvoid: Callable[[Parser[U]], Parser[VoidType]] = partial(pmap, const(Void))

lazy = lambda ff: lambda *args: ff()(*args)



from pathlib import Path; import sys; sys.path.append(str(Path(__file__).resolve().parent.parent / 'src'))

import os
import subprocess

from jax import Array
import jax.numpy as jnp

import einshard

def set_device_count(n: int) -> None:
    os.environ['JAX_PLATFORMS'] = 'cpu'
    os.environ['XLA_FLAGS'] = os.environ.get('XLA_FLAGS', '') + f' --xla_force_host_platform_device_count={n}'

def get_shard_shape(a: Array) -> tuple[int, ...]:
    return a.addressable_data(0).shape

def assert_equal(a, b):
    if a != b:
        raise AssertionError(f'{a} != {b}')

tests = [
    {
        'n_devices': 1,
        'shape': (2,),
        'expr': '... -> * ...',
        'ans': (2,),
    },
    {
        'n_devices': 16,
        'shape': (2,),
        'expr': 'a -> 16 a',
        'ans': (2,),
    },
    {
        'n_devices': 16,
        'shape': (2, 8),
        'expr': 'a b -> 2 a b*',
        'ans': (2, 1),
    },
    {
        'n_devices': 8,
        'shape': (4, 4),
        'expr': 'a b -> * a* b*',
        'ans': (2, 2),
    },
    {
        'n_devices': 16,
        'shape': (2, 8),
        'expr': '... a -> 2 ... a*',
        'ans': (2, 1),
    },
    {
        'n_devices': 16,
        'shape': (4, 16, 8),
        'expr': 'a b c -> a2* b2 c*',
        'ans': (1, 8, 4),
    },
]

def invoke_test(spec) -> None:
    set_device_count(spec['n_devices'])
    a = jnp.zeros(spec['shape'])
    a = einshard.shard(a, spec['expr'])
    assert_equal(get_shard_shape(a), spec['ans'])

def main() -> None:
    if len(sys.argv) < 2:  # no command-line arguments provided
        for i in range(len(tests)):
            result = subprocess.call([sys.executable, __file__, f'{i}'])
            if result != 0:
                print(f'❌ Test {i} failed')
            else:
                print(f'✅ Test {i} passed')
    else:
        i = int(sys.argv[1])
        invoke_test(tests[i])

if __name__ == "__main__":
    main()



from types import EllipsisType
from typing import Callable
import unicodedata

from .parsec import anyof, const, literal, many, many1, parse_eof, pchain, pjoin, pmap, pvoid, satisfy, sepby1, with_default

type ElementLeft = str | EllipsisType
type ElementRight = tuple[str | None, int, bool] | EllipsisType
type Expression = tuple[list[ElementLeft], list[ElementRight]]

is_identifier_char: Callable[[str], bool] = lambda c: unicodedata.category(c)[0] == 'L' or c == '_'
is_0_to_9: Callable[[str], bool] = lambda c: c.isdigit()
is_1_to_9: Callable[[str], bool] = lambda c: c.isdigit() and c != '0'
is_space: Callable[[str], bool] = lambda c: c.isspace()

parse_0_to_9 = satisfy(is_0_to_9, 'digit 0-9')
parse_1_to_9 = satisfy(is_1_to_9, 'digit 1-9')
parse_integer = pmap(int, pjoin(pchain(parse_1_to_9, pjoin(many(parse_0_to_9)))))

parse_identifier_char = satisfy(is_identifier_char, 'identifier')
parse_identifier = pjoin(many1(parse_identifier_char))

parse_space = satisfy(is_space, 'space')
parse_spaces = many1(parse_space)
parse_spaces_optional = pvoid(many(parse_space))

parse_right_arrow = pvoid(literal('->'))
parse_ellipsis = pmap(const(...), literal('...', desc='ellipsis'))
parse_asterisk = pmap(const(True), literal('*'))

parse_identifier_optional = with_default(parse_identifier, default=None)
parse_integer_optional = with_default(parse_integer, default=1)
parse_asterisk_optional = with_default(parse_asterisk, default=False)

parse_element_left = anyof(parse_identifier, parse_ellipsis)
parse_element_right = anyof(
    pchain(parse_identifier_optional, parse_integer, parse_asterisk_optional),
    pchain(parse_identifier, parse_integer_optional, parse_asterisk_optional),
    pchain(parse_identifier_optional, parse_integer_optional, parse_asterisk),
    parse_ellipsis,
)
parse_elements_left = sepby1(parse_element_left, parse_spaces)
parse_elements_right = sepby1(parse_element_right, parse_spaces)
parse_expression = pchain(
    parse_spaces_optional,
    parse_elements_left,
    parse_spaces_optional,
    parse_right_arrow,
    parse_spaces_optional,
    parse_elements_right,
    parse_spaces_optional,
    parse_eof,
)

# print(pchain(literal('123 '), parse_element_left)('123 .a..', 0))
# print(parse_element_left('a', 0))
# print(parse_element_left('...', 0))
# print(parse_element_right('...', 0))
# print(parse_expression('a ... b -> b ... a1', 0))



from typing import Generic, TypeVar

T = TypeVar('T')
E = TypeVar('E')

type Either[T, E] = success[T] | fail[E]

class success(Generic[T]):
    def __init__(self, value: T):
        self.value = value

    def __repr__(self):
        return f'success({self.value!r})'

    def is_success(self) -> bool:
        return True

    def get(self) -> T:
        return self.value

    def __eq__(self, other):
        if not isinstance(other, success):
            return False
        return self.value == other.value

class fail(Generic[E]):
    def __init__(self, error: E):
        self.error = error

    def __repr__(self):
        return f'fail({self.error!r})'

    def is_success(self) -> bool:
        return False

    def get(self) -> E:
        return self.error

    def __eq__(self, other):
        if not isinstance(other, fail):
            return False
        return self.error == other.error

# def might_fail(n: int) -> Either[str, int]:
#     if n > 0:
#         return Success('Positive number')
#     else:
#         return Fail(-1)

# result = might_fail(-1)
# print(result)



from .Either import Either, success, fail
from .Void import VoidType, Void
from .parsec import ParseResult, ParseResultFailed, ParseResultSucceeded, Parser, ParserFailed, \
    ParserSucceeded, anyof, const, lazy, literal, many, many1, parse_eof, pchain, pjoin, pmap, \
    pselect, pvoid, satisfy, sepby1, with_default, with_description



from typing import NoReturn

class VoidType:
    '''A private class to implement the singleton pattern.'''
    def __repr__(self):
        return 'Void'

Void = VoidType()

def _void_type_new(cls) -> NoReturn:
    raise RuntimeError('VoidType cannot be instantiated directly')

def _void_type_init_subclass(cls, **kwargs) -> NoReturn:
    raise RuntimeError('Subclassing VoidType is not allowed')

VoidType.__new__ = _void_type_new
VoidType.__init_subclass__ = _void_type_init_subclass



from math import prod

import jax
from jax import Array
from jax.experimental import mesh_utils
from jax.sharding import Mesh, NamedSharding, PartitionSpec as P

from .parser import parse_expression

def _partition_at_ellipsis(lst: list) -> tuple[list, list]:
    idx = lst.index(...)
    l = lst[:idx]
    r = lst[idx + 1:]
    return l, r

def sharding(expression: str, *, n_dims: int | None = None) -> NamedSharding:
    '''
    Get sharding from einshard expression.

    Args:
        expression (str): The einshard expression.
        n_dims (int | None): The number of dimensions of the array to be sharded. This argument must be provided if ellipsis is used in the einshard expression.

    Returns:
        jax.sharding.NamedSharding: The :class:`jax.sharding.Sharding` object corresponding to the given expression.
    '''
    n_devices = jax.device_count()

    res = parse_expression(expression, 0)
    if not res.is_success():
        idx, desc = res.error
        raise ValueError(f'Cannot parse einshard expression "{expression}", expected {desc} at position {idx}.')
    _, (elements_left, elements_right) = res.value

    n_left_ellipses = sum(element_left is ... for element_left in elements_left)
    n_right_ellipses = sum(element_right is ... for element_right in elements_right)
    assert n_left_ellipses == n_right_ellipses and n_left_ellipses <= 1

    if n_left_ellipses == 0:
        if n_dims is not None:
            assert n_dims == len(elements_left)
    else:  # == 1
        assert n_dims is not None
        n_dims_elided = n_dims - len(elements_left) + 1
        axis_names_for_left_augmented = [f'?{i}' for i in range(n_dims_elided)]
        axis_names_for_right_augmented = [(identifier, 1, False) for identifier in axis_names_for_left_augmented]  # 1: `sharding_number`, False: `is_proportional`

        elements_left_left, elements_left_right = _partition_at_ellipsis(elements_left)
        elements_left = [*elements_left_left, *axis_names_for_left_augmented, *elements_left_right]

        elements_right_left, elements_right_right = _partition_at_ellipsis(elements_right)
        elements_right = [*elements_right_left, *axis_names_for_right_augmented, *elements_right_right]

        # print(elements_left)
        # print(elements_right)

    sharding_numbers_fixed = [sharding_number for _, sharding_number, is_proportional in elements_right if not is_proportional]
    sharding_numbers_proportional = [sharding_number for _, sharding_number, is_proportional in elements_right if is_proportional]

    if not sharding_numbers_proportional:
        sharding_ratio = 1  # can be of whatever value because it will not be used in this case
    else:
        n_devices_needed_for_fixed = prod(sharding_numbers_fixed)
        n_devices_needed_for_proportional_base = prod(sharding_numbers_proportional)
        n_sharded_axes_proportional = len(sharding_numbers_proportional)

        assert n_devices % n_devices_needed_for_fixed == 0
        n_devices_available_proportional = n_devices // n_devices_needed_for_fixed

        assert n_devices_available_proportional % n_devices_needed_for_proportional_base == 0
        sharding_ratio_full = n_devices_available_proportional // n_devices_needed_for_proportional_base
        sharding_ratio = sharding_ratio_full ** (1 / n_sharded_axes_proportional)
        assert sharding_ratio.is_integer()
        sharding_ratio = int(sharding_ratio)

    mesh_shape = [sharding_number * (1 if not is_proportional else sharding_ratio) for _, sharding_number, is_proportional in elements_right]
    axis_names = tuple(f'a{i}' for i, _ in enumerate(elements_right))
    d = {identifier: i for i, (identifier, _, _) in enumerate(elements_right) if identifier is not None}
    partition_spec = tuple(f'a{d[element_left]}' for element_left in elements_left)

    # print(mesh_shape)
    # print(axis_names)
    # print(partition_spec)

    devices = mesh_utils.create_device_mesh(mesh_shape)
    mesh = Mesh(devices, axis_names=axis_names)
    sharding_ = NamedSharding(mesh, P(*partition_spec))
    return sharding_

def shard(arr: Array, expression: str) -> Array:
    '''
    Shards a :class:`jax.Array` according to the given einshard expression.

    Args:
        arr (jax.Array): The array to be sharded.
        expression (str): The einshard expression.

    Returns:
        jax.Array: The sharded array.
    '''
    sharding_ = sharding(expression, n_dims=len(arr.shape))
    arr = jax.make_array_from_callback(arr.shape, sharding_, lambda idx: arr[idx])
    return arr
